/* 
   Copyright 2003 Savonet team
   Some parts from oggenc/oggdec, (c) 2000-2002 Michael Smith
   Some parts from ices2 (c) 2001 Michael Smith <msmith@labyrinth.net.au>
   See COPYRIGHT file for details.

   This file is part of Ocaml-vorbis.
   
   Ocaml-vorbis is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   Ocaml-vorbis is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with Ocaml-vorbis; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/**
 * Libvorbis bindings for OCaml.
 *
 * @author Samuel Mimram, Julien Cristau
 */

/* $Id: vorbis_stubs.c,v 1.25 2004/06/29 22:56:47 smimram Exp $ */

/*
  TODO:
  - handle errors
  - advanced encoder options
*/

#include <caml/alloc.h>
#include <caml/callback.h>
#include <caml/fail.h>
#include <caml/memory.h>
#include <caml/misc.h>
#include <caml/mlvalues.h>

#include <vorbis/vorbisfile.h>
#include <vorbis/vorbisenc.h>
#include <vorbis/codec.h>

#include <string.h>
#include <stdio.h>
#include <time.h>

#include "utf8.h"

struct myvorbis__encoder_t
{
  vorbis_info vi;
  vorbis_dsp_state vd;
  vorbis_block vb;
  ogg_stream_state os;
  ogg_page og;
  ogg_packet op;

  int in_channels;
  int in_samplesize;
  int in_bigendian;
  int serial;
};

typedef struct myvorbis__encoder_t myvorbis_encoder_t;

static myvorbis_encoder_t* ve_of_block(value block)
{
  return (myvorbis_encoder_t*)Field(block, 1);
}

static void finalize_myvorbis_encoder_t(value v_e)
{
  myvorbis_encoder_t *ve = ve_of_block(v_e);

  ogg_stream_clear(&ve->os);
  vorbis_block_clear(&ve->vb);
  vorbis_dsp_clear(&ve->vd);
  vorbis_info_clear(&ve->vi);

  free(ve);
}

static int int_of_option(value v)
{
  if(Is_block(v))
    return Int_val(Field(v, 0));
  else
    return -1;
}

/*
static double double_of_option(value v)
{
  if(Is_block(v))
    return Double_val(Field(v, 0));
  else
    return -1;
}
*/

static char* string_of_option(value v)
{
  if(Is_block(v))
    return String_val(Field(v, 0));
  else
    return NULL;
}

static value copy_buffer(char const *b, int len)
{
  CAMLlocal1(res);
  res = alloc_string(len);
  memmove(String_val(res), b, len);
  return res;
}

static char* strncat_safe(char* dest, char* src, int pos, int len) {
  dest = (char*)realloc((void*)dest, sizeof(char)*(pos+len));
  memcpy(dest+pos, src, len);
  return dest;
}

static void add_tag(vorbis_comment *vc, char *name, char *val)
{
  char *utf8;
  if(utf8_encode(val, &utf8) >= 0)
    {
      if(name == NULL)
	vorbis_comment_add(vc, utf8);
      else
	vorbis_comment_add_tag(vc, name, utf8);
      free(utf8);
    }
  else
    raise_with_arg(*caml_named_value("vorbis_exn_utf8_failure"), copy_string(val));
}

static void build_comments(vorbis_comment *vc,
			   char *artist, char *album, char *title, char *tracknum,
			   char *date, char *genre)
{
  vorbis_comment_init(vc);

  if (title)
    add_tag(vc, "title", title);
  if (artist)
    add_tag(vc, "artist", artist);
  if (genre)
    add_tag(vc, "genre", genre);
  if (date)
    add_tag(vc, "date", date);
  if (album)
    add_tag(vc, "album", album);
  if (tracknum)
    add_tag(vc, "tracknumber", tracknum);
  add_tag(vc, "encoder", "ocaml-vorbis by the savonet team (http://savonet.sf.net)");
}


CAMLprim value ocaml_vorbis_get_header_frame(myvorbis_encoder_t * ve, value title, value artist, value genre, value date, value album, value tracknum)
{
  CAMLparam5(title, artist, genre, date, album);
  CAMLxparam1(tracknum);

  char *ret = NULL;
  int ret_len = 0;
  CAMLlocal1(s);

  /* Now, build the three header packets and send through to the stream
     output stage (but defer actual file output until the main encode loop) */

  {
    ogg_packet header_main;
    ogg_packet header_comments;
    ogg_packet header_codebooks;
    vorbis_comment comments;

    build_comments(&comments, string_of_option(artist), string_of_option(album), string_of_option(title), string_of_option(tracknum), string_of_option(date), string_of_option(genre));
    comments.vendor = "ocaml-vorbis";

    /* Build the packets */
    vorbis_analysis_headerout(&ve->vd, &comments,
			      &header_main, &header_comments, &header_codebooks);

    /* And stream them out */
    ogg_stream_packetin(&ve->os, &header_main);
    ogg_stream_packetin(&ve->os, &header_comments);
    ogg_stream_packetin(&ve->os, &header_codebooks);

    while(ogg_stream_flush(&ve->os, &ve->og))
      {
	ret = (char*)realloc(ret, ret_len + (&ve->og)->header_len + (&ve->og)->body_len);
	memcpy(ret + ret_len, (&ve->og)->header, (&ve->og)->header_len);
	memcpy(ret + ret_len + (&ve->og)->header_len, (&ve->og)->body, (&ve->og)->body_len);
	ret_len += (&ve->og)->header_len + (&ve->og)->body_len;
      }
  }

  s = copy_buffer(ret, ret_len);
  free(ret);
  CAMLreturn(s);
}

/* TODO: check parameters consistency */
CAMLprim value ocaml_vorbis_create_encoder(value params, value title, value artist, value genre, value date, value album, value tracknum)
{
  CAMLparam5(params, title, artist, genre, date);
  CAMLxparam2(album, tracknum);
  CAMLlocal3(encoder,string,block);

  myvorbis_encoder_t *ve;
  int quality_set = 0;

  int bitrate = int_of_option(Field(params, 0));
  int min_bitrate = int_of_option(Field(params, 1));
  int max_bitrate = int_of_option(Field(params, 2));
  float quality = Double_val(Field(params, 3));
  int channels = Int_val(Field(params, 4));
  int sample_freq = int_of_option(Field(params, 5));
  int managed = Bool_val(Field(params, 6));
  int in_channels = Int_val(Field(params, 7));
  int in_samplefreq = Int_val(Field(params, 8));
  int in_samplesize = Int_val(Field(params, 9));
  int in_bigendian = Int_val(Field(params, 10));

  int serialno;
  srand(time(NULL));
  serialno = rand();

  /* TODO: other checks */
  if(sample_freq < 8000 || sample_freq > 48000)
    raise_constant(*caml_named_value("vorbis_exn_invalid_sample_freq"));
  if(channels != 1 && channels != 2 )
    raise_constant(*caml_named_value("vorbis_exn_invalid_channels"));

  ve = malloc(sizeof(myvorbis_encoder_t));

  ve->in_channels = in_channels;
  ve->in_samplesize = in_samplesize;
  ve->in_bigendian = in_bigendian;
  ve->serial = serialno;

  /* If we had no quality or bitrate spec at all from the user, use
     the default quality with no management */
  if (bitrate < 0 && min_bitrate < 0 && max_bitrate < 0)
    quality_set = 1;

  /* Have vorbisenc choose a mode for us */
  vorbis_info_init(&ve->vi);
  if(quality_set > 0)
    {
      if (vorbis_encode_setup_vbr(&ve->vi, channels, sample_freq, quality))
	{
	  vorbis_info_clear(&ve->vi);
	  raise_constant(*caml_named_value("vorbis_exn_invalid_quality"));
	}

      /* Do we have optional hard quality restrictions? */
      if (max_bitrate > 0 || min_bitrate > 0)
	{
	  struct ovectl_ratemanage_arg ai;
	  
	  vorbis_encode_ctl(&ve->vi, OV_ECTL_RATEMANAGE_GET, &ai);  
	  ai.bitrate_hard_min = min_bitrate * 1000;
	  ai.bitrate_hard_max = max_bitrate * 1000;
	  ai.management_active = 1;
	  vorbis_encode_ctl(&ve->vi, OV_ECTL_RATEMANAGE_SET, &ai);
	}
    }
  else
    {
      if (vorbis_encode_setup_managed(&ve->vi, channels, sample_freq,
				      max_bitrate>0?max_bitrate*1000:-1, 
				      bitrate*1000,
				      min_bitrate>0?min_bitrate*1000:-1))
	{
	  vorbis_info_clear(&ve->vi);
	  raise_constant(*caml_named_value("vorbis_exn_invalid_bitrate"));
	}
    }

  if (managed && bitrate < 0)
    {
      vorbis_encode_ctl(&ve->vi, OV_ECTL_RATEMANAGE_AVG, NULL);
    }
  else if(!managed)
    {
      /* Turn off management entirely (if it was turned on). */
      vorbis_encode_ctl(&ve->vi, OV_ECTL_RATEMANAGE_SET, NULL);
    }

  /* TODO: advanced encoder options */

  vorbis_encode_setup_init(&ve->vi);
  vorbis_analysis_init(&ve->vd, &ve->vi);
  vorbis_block_init(&ve->vd, &ve->vb);

  ogg_stream_init(&ve->os, serialno);

  encoder = alloc_final(2, finalize_myvorbis_encoder_t, sizeof(myvorbis_encoder_t), 1024);
  Field(encoder, 1) = (value)ve;

  block = alloc_tuple(2);
  Field(block,0) = encoder;
  /* TODO : clean this */
  Field(block,1) = ocaml_vorbis_get_header_frame(ve, title, artist, genre, date, album, tracknum);

  CAMLreturn(block);
}

CAMLprim value ocaml_vorbis_create_encoder_byte(value* argv, int argc) {
  return (ocaml_vorbis_create_encoder(argv[0], argv[1], argv[2], argv[3], argv[4], argv[5], argv[6]));
}


/**
 * It does initialize buffer_.
 */
static int wav_read(myvorbis_encoder_t *ve, char *in, float ***buffer_, int len)
{
  int sampbyte = ve->in_samplesize / 8;
  char *buf;
  int bytes_read;
  int i, j;
  int realsamples;
  float **buffer;

  realsamples = len / (sampbyte * ve->in_channels);
  *buffer_ = vorbis_analysis_buffer(&ve->vd, realsamples);
  buffer = *buffer_;
  bytes_read = realsamples * sampbyte * ve->in_channels;
  buf = (char*)malloc(bytes_read);
  memcpy(buf, in, bytes_read);

  if(ve->in_samplesize == 8)
    {
      unsigned char *bufu = (unsigned char *)buf;
      for(i = 0; i < realsamples; i++)
	{
	  for(j=0; j < ve->in_channels; j++)
	    {
	      buffer[j][i] = ((int)(bufu[i * ve->in_channels + j]) - 128) / 128.0f;
	    }
	}
    }
  else
    {
      if(!ve->in_bigendian)
	{
	  for(i = 0; i < realsamples; i++)
	    {
	      for(j=0; j < ve->in_channels; j++)
		{
		  buffer[j][i] = ((buf[i * 2 * ve->in_channels + 2 * j + 1] << 8) |
				  (buf[i * 2 * ve->in_channels + 2 * j] & 0xff)) / 32768.0f;
		}
	    }
	}
      else
	{
	  for(i = 0; i < realsamples; i++)
	    {
	      for(j=0; j < ve->in_channels; j++)
		{
		  buffer[j][i] = ((buf[i * 2 * ve->in_channels + 2 * j] << 8) |
				  (buf[i * 2 * ve->in_channels + 2 * j + 1] & 0xff)) / 32768.0f;
		}
	    }
	}
    }
  free(buf);

  return realsamples;
}


CAMLprim value ocaml_vorbis_encode_buffer(value v_e, value buf)
{
  CAMLparam2(v_e, buf);

  myvorbis_encoder_t *ve = ve_of_block(v_e);
  float **buffer;
  int samples_read = wav_read(ve, String_val(buf), &buffer, string_length(buf));
  char *ret = NULL;
  int ret_len = 0;
  int eos = 0;
  CAMLlocal1(s);

  if(samples_read == 0)
    /* Tell the library that we wrote 0 bytes - signalling the end */
    vorbis_analysis_wrote(&ve->vd, 0);
  else
    /* Tell the library how many samples (per channel) we wrote 
       into the supplied buffer */
    vorbis_analysis_wrote(&ve->vd, samples_read);

  /* While we can get enough data from the library to analyse, one
     block at a time... */
  while(vorbis_analysis_blockout(&ve->vd, &ve->vb) == 1)
    {
      /* Do the main analysis, creating a packet */
      vorbis_analysis(&ve->vb, NULL);
      vorbis_bitrate_addblock(&ve->vb);
      
      while(vorbis_bitrate_flushpacket(&ve->vd, &ve->op)) 
	{
	  /* Add packet to bitstream */
	  ogg_stream_packetin(&ve->os, &ve->op);
	  
	  /* If we've gone over a page boundary, we can do actual output,
	     so do so (for however many pages are available) */	  
	  while(!eos)
	    {
	      if(!ogg_stream_pageout(&ve->os, &ve->og)) break;

	      ret = (char*)realloc(ret, ret_len + (&ve->og)->header_len + (&ve->og)->body_len);
	      memcpy(ret + ret_len, (&ve->og)->header, (&ve->og)->header_len);
	      memcpy(ret + ret_len + (&ve->og)->header_len, (&ve->og)->body, (&ve->og)->body_len);
	      ret_len += (&ve->og)->header_len + (&ve->og)->body_len;
	      
	      if(ogg_page_eos(&ve->og)) eos = 1;
	    }
	}
    }

  s = copy_buffer(ret, ret_len);
  free(ret);

  CAMLreturn(s);
}


struct myvorbis__dec_file_t
{
  OggVorbis_File *ovf;
  int channels;
  int sample_freq;
  int sample_size;
  int big_endian;
  int sign;
};


typedef struct myvorbis__dec_file_t myvorbis_dec_file_t;


static myvorbis_dec_file_t* dec_file_of_block(value block)
{
  return (myvorbis_dec_file_t*)Field(block, 1);
}


static void finalize_myvorbis_dec_file_t(value d_f)
{
  myvorbis_dec_file_t *df = dec_file_of_block(d_f);

  free(df);
}


CAMLprim value ocaml_vorbis_open_dec_file(value file, value params)
{
  CAMLparam2(file, params);
  CAMLlocal1(block);

  FILE *in = NULL;
  myvorbis_dec_file_t *df = (myvorbis_dec_file_t*)malloc(sizeof(myvorbis_dec_file_t));

  if(!(in = fopen(String_val(file), "rb")))
      raise_constant(*caml_named_value("vorbis_exn_could_not_open_file"));
  df->ovf = (OggVorbis_File*)malloc(sizeof(OggVorbis_File));
  df->channels = Int_val(Field(params, 0));
  df->sample_freq = Int_val(Field(params, 1));
  df->sample_size = Int_val(Field(params, 2));
  df->big_endian = Bool_val(Field(params, 3));
  df->sign = Bool_val(Field(params, 4));
  if(ov_open(in, df->ovf, NULL, 0) < 0)
    {
      fclose(in);
      free(df->ovf);
      raise_constant(*caml_named_value("vorbis_exn_not_a_vorbis_file"));
    }

  block = alloc_final(2, finalize_myvorbis_dec_file_t, sizeof(myvorbis_dec_file_t), 1000);
  Field(block, 1) = (value)df;
  CAMLreturn(block);
}

/* TODO: convert file format? */
CAMLprim value ocaml_vorbis_decode(value d_f, value buf_, value ofs_, value len_)
{
  CAMLparam4(d_f, buf_, ofs_, len_);

  myvorbis_dec_file_t *df = dec_file_of_block(d_f);
  int ret = 0;
  int bs = 0;
  int ofs = Int_val(ofs_);
  int len = Int_val(len_);
  char* buf = String_val(buf_) + ofs;

  if (ofs + len > string_length(buf_))
    raise_constant(*caml_named_value("vorbis_exn_invalid_parameters"));
  ret = ov_read(df->ovf, buf, len, df->big_endian, df->sample_size / 8, df->sign, &bs);
  if(ret == 0)
      raise_constant(*caml_named_value("exn_end_of_file"));
  if(ret < 0)
      raise_constant(*caml_named_value("vorbis_exn_hole_in_data"));

  CAMLreturn(Val_int(ret));
}


CAMLprim value ocaml_vorbis_close_dec_file(value d_f)
{
  CAMLparam1(d_f);
  myvorbis_dec_file_t* df = dec_file_of_block(d_f);

  ov_clear(df->ovf);

  CAMLreturn(Val_unit);
}

CAMLprim value ocaml_vorbis_utf8_decode(value string)
{
  CAMLparam1(string);
  char* utf8;
  if(utf8_decode(String_val(string),&utf8) >= 0)
    {
      CAMLreturn(copy_string(utf8));
    }
  raise_with_arg(*caml_named_value("vorbis_exn_utf8_failure"), string);
}

CAMLprim value ocaml_vorbis_encoder_reset(value v_e, value title, value artist, value genre, value date, value album, value tracknum) {
  CAMLparam5(v_e, title, artist, genre, date);
  CAMLxparam2(album, tracknum);

  myvorbis_encoder_t *s = ve_of_block(v_e);

  ogg_packet op;
  ogg_page og;
  CAMLlocal1(string);
  char* ret=NULL;
  int size = 0;

  /* tell libvorbis that this is the end of the stream */
  vorbis_analysis_wrote(&s->vd, 0);
  
  while(vorbis_analysis_blockout(&s->vd, &s->vb)==1)
    {
      vorbis_analysis(&s->vb, NULL);
      vorbis_bitrate_addblock(&s->vb);
      while(vorbis_bitrate_flushpacket(&s->vd, &op))
	ogg_stream_packetin(&s->os, &op);
    }

  /* flush the last pages in a buffer */
  while(ogg_stream_pageout(&s->os, &og)<=0?0:1) {
    ret = strncat_safe(ret, og.header, size, og.header_len);
    size += og.header_len;
    ret = strncat_safe(ret, og.body, size, og.body_len);
    size += og.body_len;
  }

  /* clear the encoder */
  ogg_stream_clear(&s->os);
  vorbis_block_clear(&s->vb);
  vorbis_dsp_clear(&s->vd);

  /* build the new vorbis comments */
  do {
    vorbis_comment comments;
    ogg_packet h1,h2,h3;

    build_comments(&comments, string_of_option(artist), string_of_option(album), string_of_option(title), string_of_option(tracknum), string_of_option(date), string_of_option(genre));
  
    comments.vendor = "ocaml-vorbis";

    vorbis_analysis_init(&s->vd, &s->vi);
    vorbis_block_init(&s->vd, &s->vb);
    
    ogg_stream_init(&s->os, ++(s->serial));
    vorbis_analysis_headerout(&s->vd, &comments, &h1,&h2,&h3);
    ogg_stream_packetin(&s->os, &h1);
    ogg_stream_packetin(&s->os, &h2);
    ogg_stream_packetin(&s->os, &h3);

  } while(0);

  /* add the headers to the returned buffer */
  while(ogg_stream_flush(&s->os, &s->og))
    {
      ret = (char*)realloc(ret, size + (&s->og)->header_len + (&s->og)->body_len);
      memcpy(ret + size, (&s->og)->header, (&s->og)->header_len);
      memcpy(ret + size + (&s->og)->header_len, (&s->og)->body, (&s->og)->body_len);
      size += (&s->og)->header_len + (&s->og)->body_len;
    }
  
  string = copy_buffer(ret, size);
  free(ret);
  CAMLreturn(string);
}

value ocaml_vorbis_encoder_reset_byte(value *argv, int argc)
{
  return (ocaml_vorbis_encoder_reset(argv[0], argv[1], argv[2], argv[3], argv[4], argv[5], argv[6]));
}
